# -*- coding: utf-8 -*-

# Public Domain (-) 2004-2011 The a Authors.
# See the Yatiblog UNLICENSE file for details.

r"""
================
reStructuredText
================

reStructuredText is an easy to use markup language for plain text content. This
module enhances the rendered output with a lot of prettiness and provides a
helper utility function called ``render_rst``.

  >>> text = '''
  ...
  ... This is a **test** document.
  ...
  ... '''

By default the output format of ``xhtml`` is assumed:

  >>> render_rst(text)
  u'<p>This is a <strong>test</strong> document.</p>'

You can also pass in the alternative ``format='tex'``. This produces TeX output
which should be run through an application like ``latex`` or ``pdflatex`` to get
a typeset document.

By default the input is assumed to be either a unicode object or a str object
encoded as 'utf-8'. You can specify an alternative input encoding if needed:

  >>> text = '''
  ...
  ... Price in Caf\xe9s are often in \x80
  ...
  ... '''
  ...
  >>> render_rst(text, encoding='windows-1252')
  u'<p>Price in Caf\xe9s are often in \u20ac</p>'

The output is always a unicode object which can then be decoded into other
codecs if needed.

Let's look at a sample text with a few more elements:

  >>> text = '''
  ...
  ... ==============
  ... Some Big Title
  ... ==============
  ...
  ... Yet Another Title
  ... -----------------
  ...
  ... :Author: tav
  ... :Some-Field: some value
  ...
  ... Hello *world*!
  ...
  ... '''

You can set the optional ``as_whole`` parameter to True and a complete HTML or
LaTeX output will be generated with default headers and footers:

  >>> render_rst(text, as_whole=True)
  u'...<html...<p>Hello <em>world</em>!</p>...'

As opposed to just the rendered version of the text itself:

  >>> render_rst(text)
  u'<p>Hello <em>world</em>!</p>'

If you want the data in the bibliographic fields, you can get it by specifying
the optional ``with_docinfo`` parameter:

  >>> print render_rst(text, with_docinfo=True)
  <div class="docinfo">
  <table class="docinfo" frame="void" rules="none">
  <col class="docinfo-name" />
  <col class="docinfo-content" />
  <tbody valign="top">
  <tr><th class="docinfo-name">Author:</th>
  <td>tav</td></tr>
  <tr class="field"><th class="docinfo-name">Some-Field:</th><td class="field-body">some value</td>
  </tr>
  </tbody>
  </table>
  <BLANKLINE>
  </div>
  <div class="document">
  <p>Hello <em>world</em>!</p>
  </div>

If the bibliographic data is needed as a dictionary, then the optional
``with_props`` parameter can be specified:

  >>> output, props = render_rst(text, with_props=True)

  >>> output
  u'<p>Hello <em>world</em>!</p>'

  >>> sorted(props.keys())
  [u'author', u'some-field', u'subtitle', u'title']

As can be seen, this includes the bibliographic fields as well as the extracted
text title and subtitle:

  >>> props['author']
  u'tav'

  >>> props['some-field']
  u'some value'

  >>> props['title']
  u'Some Big Title'

  >>> props['subtitle']
  u'Yet Another Title'

The HTML output has a number of typographic additions like smart quotes and em
dashes:

  >>> text = '''
  ...
  ... "This is a quote" -- Gandhi
  ...
  ... '''

  >>> render_rst(text)
  u'<p>&ldquo;This is a quote&rdquo; &mdash; Gandhi</p>'

And, finally, in addition to the default directives, a ``syntax`` directive has
been added which allows for the syntax highlighting of included source code in a
variety of languages.

  >>> text = '''
  ...
  ... This is some text
  ...
  ... .. syntax:: python
  ...
  ...     double = lambda x: 2 * x
  ...
  ... '''

  >>> print render_rst(text)
  <p>This is some text</p>
  <div class="syntax python"><pre>...<span class="k">lambda</span>...</pre></div>

"""

import re

from string import punctuation as PUNCTUATION

from docutils import nodes
from docutils.frontend import OptionParser
from docutils.parsers.rst import directives, Directive, DirectiveError, Parser
from docutils.parsers.rst.states import RSTStateMachine, state_classes
from docutils.readers.standalone import Reader
from docutils.transforms import writer_aux, universal, references, frontmatter
from docutils.transforms import misc
from docutils.writers.html4css1 import HTMLTranslator, Writer as HTMLWriter
from docutils.writers.latex2e import LaTeXTranslator, Writer as LaTexWriter
from docutils.utils import new_document

from pygments import highlight
from pygments.formatters import HtmlFormatter
from pygments.lexers import get_lexer_by_name, TextLexer

try:
    from simplejson import loads as decode_json
except ImportError:
    from json import loads as decode_json

from tavutil.io import IteratorParser

# ------------------------------------------------------------------------------
# Some Constants
# ------------------------------------------------------------------------------

HTML_VISITOR_ATTRIBUTES = (
    'head_prefix', 'head', 'stylesheet', 'body_prefix',
    'body_pre_docinfo', 'docinfo', 'body', 'body_suffix',
    'title', 'subtitle', 'header', 'footer', 'meta', 'fragment',
    'html_prolog', 'html_head', 'html_title', 'html_subtitle',
    'html_body'
    )

LATEX_VISITOR_ATTRIBUTES = (
    "head_prefix", "head", "body_prefix", "body", "body_suffix"
    )

DEFAULT_TRANSFORMS = [

    universal.Decorations,
    universal.ExposeInternals,
    universal.StripComments,

    references.Substitutions,
    references.PropagateTargets,
    frontmatter.DocTitle,
    frontmatter.SectionSubTitle,
    frontmatter.DocInfo,
    references.AnonymousHyperlinks,
    references.IndirectHyperlinks,
    references.Footnotes,
    references.ExternalTargets,
    references.InternalTargets,
    references.DanglingReferences,
    misc.Transitions,

    ]

HTML_TRANSFORMS = DEFAULT_TRANSFORMS + [

    universal.Messages,
    universal.FilterMessages,
    universal.StripClassesAndElements,

    writer_aux.Admonitions,

    ]

OPTION_PARSER = OptionParser((Parser, Reader))
HTML_OPTION_PARSER = OptionParser((Parser, Reader, HTMLWriter))
LATEX_OPTION_PARSER = OptionParser((Parser, Reader, LaTexWriter))

HTML_SETUP = ('html', HTMLTranslator, HTML_TRANSFORMS, HTML_OPTION_PARSER)
LATEX_SETUP = ('tex', LaTeXTranslator, DEFAULT_TRANSFORMS, LATEX_OPTION_PARSER)
RAW_SETUP = (None, None, DEFAULT_TRANSFORMS, OPTION_PARSER)

# ------------------------------------------------------------------------------
# Some Pre-compiled Regular Expressions
# ------------------------------------------------------------------------------

replace_toc_attributes = re.compile(
    '(?sm)<p class="topic-title(.*?)"><a name="(.*?)">(.*?)</a></p>(.*?)</div>'
    ).sub

replace_drop_shadows = re.compile(
    '(?sm)<div class="figure(.*?)<p><img(.*?)/></p>(.*?)</div>'
    ).sub

replace_abstract_attributes = re.compile('<div class="abstract topic">').sub
replace_ampersands = re.compile('&(?![^\s&]*;)').sub
replace_comments = re.compile(r'(?sm)\n?\s*<!--(.*?)\s-->\s*\n?').sub
replace_plexlinks = re.compile('(?sm)\[\[(.*?)\]\]').sub
replace_title_headings = re.compile('(?sm)<h1 class="title">(.*?)</h1>').sub
replace_table_borders = re.compile('(?sm)<table border="1" class="docutils">').sub
replace_whitespace = re.compile('[\v\f]').sub

split_html_tags = re.compile('(?sm)<(.*?)>').split

# ------------------------------------------------------------------------------
# RST Directives
# ------------------------------------------------------------------------------

def break_directive(name, arguments, options, content, lineno,
                    content_offset, block_text, state, state_machine):
    """Render an HTML break element."""

    if arguments:
        break_class = arguments[0]
    else:
        break_class = 'clear'

    raw_node = nodes.raw('', '<hr class="%s" />' % break_class, format='html')

    # return [nodes.transition()]

    return [raw_node]

break_directive.arguments = (0, 1, True)
break_directive.options = {
    'class':directives.class_option
    }
break_directive.content = False

directives.register_directive('break', break_directive)

# ------------------------------------------------------------------------------
# A Tag Directive!!
# ------------------------------------------------------------------------------

SEEN_TAGS_CACHE = None
TAG_COUNTER = None

class TagDirective(Directive):
    """Convert tags into HTML annotation blocks."""

    required_arguments = 0
    optional_arguments = 1
    final_argument_whitespace = True

    has_content = True

    def run(
        self, tag_cache={}, tag_normalised_cache={}, TODO=u'✗', DONE=u'✓',
        letters='abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ -'
        ):

        if not self.content:
            return []

        if self.arguments:
            arguments = self.arguments[0].strip().split(',')
        else:
            arguments = []

        tag_list_id = CURRENT_PLAN_ID or 'list'

        output = []; add = output.append
        implicit_tags = {}
        tag_id = None

        def add_tag(
            ori_tag, tag_span, cache=tag_normalised_cache, add=add,
            implicit_tags=implicit_tags,
            done_tags=CURRENT_PLAN_SETTINGS.get('done', []),
            todo_tags=CURRENT_PLAN_SETTINGS.get('todo', []),
            ):
            add(tag_span)
            norm_tag = cache[ori_tag]
            if norm_tag in done_tags:
                implicit_tags['done'] = True
            if norm_tag in todo_tags:
                implicit_tags['todo'] = True

        for tag in arguments:

            if tag in tag_cache:
                add_tag(tag, tag_cache[tag])
                continue

            ori_tag = tag
            tag = tag.strip().rstrip(u',').strip()

            if not tag:
                continue

            if tag.startswith('id:'):
                tag_id = tag[3:]
                continue

            if tag.startswith('@'):
                tag_class = u'-'.join(tag[1:].lower().split())
                tag_type = 'zuser'
                tag_text = tag
            elif tag.startswith('#'):
                tag_class = u'-'.join(tag[1:].lower().split())
                tag_type = '2'
                tag_text = tag
            elif ':' in tag:
                tag_split = tag.split(':', 1)
                if len(tag_split) == 2:
                    tag_type, tag_text = tag_split
                else:
                    tag_type = tag_split[0]
                    tag_text = u""
                tag_type = tag_type.strip().lower()
                if tag_type == 'dep' and ':' not in tag_text:
                    tag_class = u'dep-%s-%s' % (CURRENT_PLAN_ID, tag_text)
                else:
                    tag_class = u'-'.join(tag.replace(':', ' ').lower().split())
            else:
                tag_class = u'-'.join(tag.lower().split())
                tag_type = '1'
                tag_text = tag.upper()

            if ':' in tag:
                lead = tag.split(':', 1)[0]
                tag_name = '%s:%s' % (lead.lower(), tag_text)
            else:
                tag_name = tag_text

            tag_span = (
                u'<span class="tag tag-type-%s tag-val-%s" tagname="%s" '
                 'tagnorm="%s">%s</span> ' %
                (tag_type, tag_class, tag_name, tag.lower(), tag_text)
                )

            tag_cache[ori_tag] = tag_span
            tag_normalised_cache[ori_tag] = tag_class
            add_tag(ori_tag, tag_span)

        content = u'\n'.join(self.content)
        if content.startswith(TODO):
            if self.content:
                self.content[0] = self.content[0].replace(TODO, u'*%s*' % TODO, 1)
            if 'todo' not in implicit_tags:
                add(
                    u'<span class="tag tag-type-1 tag-val-todo" '
                     'tagname="TODO" tagnorm="todo">TODO</span> '
                    )
        elif content.startswith(DONE):
            if self.content:
                self.content[0] = self.content[0].replace(DONE, u'*%s*' % DONE, 1)
            if 'done' not in implicit_tags:
                add(
                    u'<span class="tag tag-type-1 tag-val-done" '
                     'tagname="DONE" tagnorm="done">DONE</span> '
                    )

        output.sort()

        if not tag_id:
            tag_id = '-'.join(''.join([
                char for char in
                content.split('\n')[0].strip()[1:].strip()
                if char in letters
                ]).split()).lower()

        if not tag_id:
            global TAG_COUNTER
            TAG_COUNTER = TAG_COUNTER + 1
            tag_id = 'temp-%s' % TAG_COUNTER

        tag_id = 'planitem-%s' % tag_id

        if tag_id in SEEN_TAGS_CACHE:
            raise DirectiveError(2, "The tag id %r has already been used!" % tag_id)

        SEEN_TAGS_CACHE.add(tag_id)

        if not output:
            pass
            # add(u'<span class="tag tag-untagged"></span>')

        add(u'<a class="tag-link" href="#%s">#</a>' % tag_id)
        output.insert(
            0, (u'<div class="tag-segment" id="%s-tags">' % tag_id)
            )

        output.append(u'</div>')

        tag_info = nodes.raw('', u''.join(output), format='html')
        tag_content_container = nodes.container(
            ids=['%s-content' % tag_id]
            )

        self.state.nested_parse(
            self.content, self.content_offset, tag_content_container
            )

        prefix = nodes.raw(
            '', u'<div id="%s" class="tag-content">' % tag_id, format='html'
            )
        suffix = nodes.raw('', u'</div>', format='html')

        return [prefix, tag_content_container, tag_info, suffix]

directives.register_directive('tag', TagDirective)

# ------------------------------------------------------------------------------
# Plan Directive!
# ------------------------------------------------------------------------------

CURRENT_PLAN_ID = None
CURRENT_PLAN_SETTINGS = {}

def plan_directive(name, arguments, options, content, lineno,
                   content_offset, block_text, state, state_machine):
    """Setup for tags relating to a plan file."""

    global CURRENT_PLAN_ID
    global CURRENT_PLAN_SETTINGS

    if not CURRENT_PLAN_ID:
        raw_node = nodes.raw(
            '',
            '<div id="plan-container"></div>'
            '<script type="text/javascript" src="js/plan.js"></script>'
            '<hr class="clear" />',
            format='html'
            )
    else:
        raw_node = nodes.raw('', '', format='html')

    content = '\n'.join(content)
    if content:
        CURRENT_PLAN_SETTINGS = decode_json(content)
    else:
        CURRENT_PLAN_SETTINGS = {}

    CURRENT_PLAN_ID = arguments[0]

    return [raw_node]

plan_directive.arguments = (1, 0, True)
plan_directive.options = {}
plan_directive.content = True

directives.register_directive('plan', plan_directive)

# ------------------------------------------------------------------------------
# Convert plain code snippets to funky HTML
# ------------------------------------------------------------------------------

SYNTAX_FORMATTER = HtmlFormatter(cssclass='syntax', lineseparator='<br/>')

def syntax_directive(name, arguments, options, content, lineno,
                     content_offset, block_text, state, state_machine):
    """Prettify <syntax> snippets into marked up HTML blocks."""

    try:
        lexer_name = arguments[0]
        lexer = get_lexer_by_name(lexer_name)
    except ValueError:
        lexer_name = 'txt'
        lexer = TextLexer()

    formatter = HtmlFormatter(
        cssclass='syntax %s' % lexer_name, lineseparator='<br/>'
        )

    return [nodes.raw(
        '',
        highlight(u'\n'.join(content), lexer, formatter),
        format='html'
        )]

syntax_directive.arguments = (1, 0, False)
syntax_directive.options = {
    'format': directives.unchanged
    }
syntax_directive.content = True

directives.register_directive('syntax', syntax_directive)

def doctest2html(content):
    """Convert doctest strings to CSS'd HTML."""

    out = []

    for line in content.splitlines():
        if line.startswith('&gt;&gt;&gt;') or line.startswith('...'):
            line = '<span class="doctest-input">' + line + '</span>'
        elif line:
            line = '<span class="doctest-output">' + line + '</span>'
        out.append(line)

    return '\n'.join(out)

# ------------------------------------------------------------------------------
# Pretty Typographical Syntax Converter
# ------------------------------------------------------------------------------

def convert(content):
    """Convert certain characters to prettier typographical syntax."""

    # Remember: the order of the replacements matter...
    content = content.replace(
        '&quot;', '"').replace(
        ' -->', 'HTML-COMMENT-ELEMENT-CLOSE').replace(
        '-&gt;', '&rarr;').replace(
        '<-', '&larr;').replace(
        '---', '&ndash;').replace(
        '--', '&mdash;').replace(
        '<<', '&laquo;').replace(
        '>>', '&raquo;').replace(
        '(C)','&copy;').replace(    # hmz, why am i promoting ipr? ;p
        '(c)','&copy;').replace(
        '(tm)','&trade;').replace(
        '(TM)','&trade;').replace(
        '(r)','&reg;').replace(
        '(R)','&reg;').replace(
        '...', '&#8230;').replace(
        'HTML-COMMENT-ELEMENT-CLOSE', ' -->')

    icontent = IteratorParser(content)
    content = []

    _scurly = _dcurly = False
    _space = True
    _apply = False

    index = 0
    prev = ''

    while True:

        try:
            char = icontent.next()
        except StopIteration:
            break

        if not (_scurly or _dcurly) and _space:
            if char == "'":
                _scurly = index + 1
            elif char == '"':
                _dcurly = index + 1

        if _scurly and (_scurly != index + 1) and char == "'" and prev != '\\':
            try:
                n = icontent.next()
                if n in PUNCTUATION or n.isspace():
                    _apply = True
                icontent.push(n)
            except:
                _apply = True
            if _apply:
                content[_scurly - 1] = '&lsquo;'
                char = '&rsquo;'
                _scurly = False
            _apply = False
        
        if _dcurly and (_dcurly != index + 1) and char == '"' and prev != '\\':
            try:
                n = icontent.next()
                if n in PUNCTUATION or n.isspace():
                    _apply = True
                icontent.push(n)
            except:
                _apply = True
            if _apply:
                content[_dcurly - 1] = '&ldquo;'
                char = '&rdquo;'
                _dcurly = False
            _apply = False

        content.append(char)
        prev = char
        index += 1
        
        if char.isspace():
            _space = True
        else:
            _space = False

    content = ''.join(content).replace('"', '&quot;')
    content = replace_plexlinks(render_plexlink, content)

    # perhaps === heading === stylee ?

    return content

# ------------------------------------------------------------------------------
# The Meta Prettifier Function Which Calls The Above Ones
# ------------------------------------------------------------------------------

def escape_and_prettify(content):
    """Escape angle brackets appropriately and prettify certain blocks."""

    # Our markers and our output gatherer.
    _literal_block = _element = _content = False
    output = []

    for i, block in enumerate(split_html_tags(content.strip())):

        # We setup the state
        if i % 2:

            _content = False
            _element = True

            if _literal_block and (_literal_block[1] == block):
                _literal_block = False

            if block == 'tt class="literal"':
                _literal_block = ('tt', '/tt')
            elif block in ['pre',
                           'pre class="literal-block"',
                           'pre class="last literal-block"',
                           'pre class="code"']:
                _literal_block = ('pre', '/pre')
            elif block in ['span class="pre"']:
                _literal_block = ('span', '/span')
            elif block == 'pre class="doctest-block"':
                _literal_block = ('doctest', '/pre')
            elif block == 'style type="text/css"':
                _literal_block = ('style type="text/css"', '/style')
            elif block == 'script type="text/javascript"':
                _literal_block = ('script type="text/javascript"', '/script')
            elif block.startswith('pre class="'):
                _literal_block = ('pre', '/pre')

        else:

            _content = True
            _element = False

        # We do different things based on the state.
        if _element:
            output.append('<' + block + '>')
        elif _content:
            if _literal_block:
                if _literal_block[0] == 'doctest':
                    output.append(doctest2html(block))
                else:
                    output.append(block)
            else:
                output.append(convert(block))

    output = ''.join(output)

    # Gah!
    output = output.replace('<<', '&lt;<')

    # Praise be to them negative lookahead regex thingies.
    output = replace_ampersands('&amp;', output)

    return output

# ------------------------------------------------------------------------------
# Some Utility Functions
# ------------------------------------------------------------------------------

def render_drop_cap(content):
    """Render the first character as a drop capital"""

    content = content.groups()[0]

    if content:
        if len(content) >= 2:
            return '<p><span class="dropcap">' + content[0] + \
                   '</span>'  + content[1:]

    return '<p></p>'

def render_plexlink(content):
    """Render [[plexlinks]]."""

    name = content.groups()[0]

    if '|' in name:
        name, linkname = name.split('|', 1)
    else:
        linkname = name

    return u'<a href="%s.html">%s</a>' % (u'-'.join(name.split()), linkname)

# ------------------------------------------------------------------------------
# Parse The :properties: Included In A Document
# ------------------------------------------------------------------------------

def parse_headers(source_lines, props, toplevel=False):
    """Parse the metadata stored in the content headers."""

    new_data = []; out = new_data.append
    iterative_source = IteratorParser(source_lines)

    for line in iterative_source:

        if line.startswith(':') and line.find(':', 2) != -1:

            marker = line.find(':', 2)
            prop, value = line[1:marker], line[marker+1:].strip()

            if prop.lower().startswith('x-'):
                _strip_prop = True
            else:
                _strip_prop = False

            if not _strip_prop:
                out(line)

            while True:
                try:
                    line = iterative_source.next()
                    if not _strip_prop:
                        out(line)
                except:
                    break
                sline = line.strip()
                if sline and not (sline.startswith(':') and \
                                  sline.find(':', 2) != -1):
                    value += ' ' + sline
                else:
                    iterative_source.push(line)
                    if not _strip_prop:
                        del new_data[-1]
                    break

            prop = prop.lower()

            if prop in props:
                oldvalue = props[prop]
                if isinstance(oldvalue, list):
                    oldvalue.append(value)
                else:
                    props[prop] = [oldvalue, value]
            else:
                props[prop] = value

        else:
            out(line)

    # re.sub('(?sm)\[\[# (.*?)\]\]', render_includes, content)

    return new_data, props

# ------------------------------------------------------------------------------
# Our Core Renderer
# ------------------------------------------------------------------------------

def render_rst(
    source, format='xhtml', encoding='utf-8', with_props=False,
    with_docinfo=False, as_whole=False
    ):
    """Return the rendered ``source`` with optional extracted properties."""

    global SEEN_TAGS_CACHE, TAG_COUNTER, CURRENT_PLAN_ID
    SEEN_TAGS_CACHE = set()
    TAG_COUNTER = 0
    CURRENT_PLAN_ID = None

    if format in ('xhtml', 'html'):
        format, translator, transforms, option_parser = HTML_SETUP
    elif format in ('tex', 'latex'):
        format, translator, transforms, option_parser = LATEX_SETUP
    elif format == 'raw':
        format, translator, transforms, option_parser = RAW_SETUP
    else:
        raise ValueError("Unknown format: %r" % format)

    settings = option_parser.get_default_values()
    settings._update_loose({
        'footnote_references': 'superscript', # 'mixed', 'brackets'
        'halt_level': 6,
        'trim_footnote_reference_space': 1,
        })

    document = new_document('[dynamic-text]', settings)

    if not isinstance(source, unicode):
        source = unicode(source, encoding)

    source = replace_whitespace(' ', source)
    source_lines = [s.expandtabs(4).rstrip() for s in source.splitlines()]

    if with_props:
        source_lines, props = parse_headers(source_lines, {}, True)

    document.reporter.attach_observer(document.note_parse_message)

    RSTStateMachine(
        state_classes=state_classes,
        initial_state='Body',
        ).run(source_lines, document)

    document.reporter.detach_observer(document.note_parse_message)
    document.current_source = document.current_line = None

    document.transformer.add_transforms(transforms)
    document.transformer.apply_transforms()

    if not format:
        return unicode(document)

    visitor = translator(document)
    document.walkabout(visitor)

    # see HTML_VISITOR_ATTRIBUTES/LATEX_VISITOR_ATTRIBUTES to see other attrs

    if as_whole:
        output = visitor.astext()
    else:
        if format == 'html' and with_docinfo:
            output = (
                u'<div class="docinfo">\n%s\n</div>\n<div class="document">\n%s</div>'
                % (u''.join(visitor.docinfo), u''.join(visitor.body))
                )
        else:
            output = u''.join(visitor.body)

    # Post RST-Conversion Prosessing
    if format == 'html':

        # [[plexlinks]]
        # output = re.sub(
        #     '(?sm)\[\[(.*?)\]\]',
        #     render_plexlink,
        #     output)

        # syntax highlighting for kode snippets
        # output = re.sub(
        #     '(?sm)<p>(?:\s)?&lt;code class=&quot;(.*?)&quot;&gt;(?::)?</p>(?:\n<blockquote>)?\n<pre class="literal-block">(.*?)</pre>(?:\n</blockquote>)?\n<p>(?:\s)?&lt;/code&gt;</p>',
        #     code2html,
        #     output)

        # Support for embedding html into RST documents and prettification.
        output = escape_and_prettify(output)

        # TOC href ID and div adder.
        output = replace_toc_attributes(
            '<p class="topic-title\\1"><a name="\\2"></a><span '
            'id="document-toc">\\3</span></p>\n<div id="documen'
            't-toc-listing">\\4</div></div>',
            output)

        # Inserting an "#abstract" ID.
        output = replace_abstract_attributes(
            r'<div id="abstract" class="abstract topic">',
            output)

        # footnote refs looking a bit too superskripted
        # output = re.sub(
        #     '(?sm)<a class="footnote-reference" (.*?)><sup>(.*?)</sup></a>',
        #     r'<a class="footnote-reference" \1>\2</a>',
        #     output)

        # Drop shadow wrappers for figures.
        output = replace_drop_shadows(
            '<div class="figure\\1<div class="wrap1"><div class="wrap2">'
            '<div class="wrap3"><img\\2/></div></div></div>\\3</div>',
            output)

        # @/@ reinstate this? -- name="" no no
        # output = re.sub(r'<a name="table-of-contents"></a>', '', output)
        # output = re.sub(r'<a (.*?) name="(.*?)">', r'<a \1>', output)

        # get rid of <p>around floating images</p>
        # output = re.sub(
        #     '(?sm)<p><img (.*?) class="float-(.*?) /></p>',
        #     r'<img \1 class="float-\2 />',
        #     output)

        # niser <hr />
        # output = re.sub(
        #     '<hr />',
        #    r'<hr noshade="noshade" />',
        #    output)

        # drop cap them first letters
        # output = re.sub(
        #     '(?sm)<p>(.*?)</p>',
        #     render_drop_cap,
        #     output, count=1)

        # Strip out comments.
        output = replace_comments('', output)

        # Strip out title headings.
        output = replace_title_headings('', output)

        # Strip out border="1".
        output = replace_table_borders(r'<table class="docutils">', output)

        # Pad out tick/cross marks.
        output = output.replace(u'<p>✓ ', u'<p>✓ &nbsp; ')
        output = output.replace(u'<p>✗ ', u'<p>✗ &nbsp; ')

    if with_props:
        if format == 'html':
            props.setdefault(u'title', visitor.title and visitor.title[0] or u'')
            props.setdefault(
                u'subtitle', visitor.subtitle and visitor.subtitle[0] or u''
                )
        return output, props

    return output

    # You can do the above by using ``docutils.core`` -- pub = Publisher() --
    # but it's a pretty inefficient way of going about converting to HTML/LaTeX.
    #
    # Anyways, speaking of pubs...
    #
    # --------------------------------------------------------------------------
    #
    # A man goes into a pub, and the barmaid asks what he wants.
    #
    # "I want to bury my face in your cleavage and lick the sweat from between
    # your tits" he says.
    #
    # "You dirty git!" shouts the barmaid, "get out before I fetch my husband."
    #
    # The man apologises and promises not to repeat his gaffe.
    #
    # The barmaid accepts this and asks him again what he wants.
    #
    # "I want to pull your pants down, spread yoghurt between the cheeks of
    # your arse and lick it all off" he says.
    #
    # "You dirty filthy pervert. You're banned! Get out!" she storms.
    #
    # Again the man apologies and swears never ever to do it again.
    #
    # "One more chance" says the barmaid.
    #
    # "Now what do you want?" "I want to turn you upside down, fill your fanny
    # with Guinness, and then drink every last drop."
    #
    # The barmaid is furious at this personal intrusion, and runs upstairs to
    # fetch her husband, who's sitting quietly watching the telly.
    #
    # "What's up, Love?" he asks.
    #
    # "There's a man in the bar who wants to put his head between my tits and
    # lick the sweat off" she says.
    #
    # "I'll kill him. Where is he?" storms the husband.
    #
    # "Then he said he wanted to pour yoghurt down between my arse cheeks and
    # lick it off" she screams.
    #
    # "Right. He's dead!" says the husband, reaching for a baseball bat.
    #
    # "Then he said he wanted to turn me upside down, fill my fanny with
    # Guinness and then drink it all" she cries.
    #
    # The husband puts down his bat and returns to his armchair, and switches
    # the telly back on.
    #
    # "Aren't you going to do something about it?" she cries hysterically.
    #
    # "Look love -- I'm not messing with someone who can drink 12 pints of
    # Guinness..."



# Public Domain (-) 2004-2011 The Yatiblog Authors.
# See the Yatiblog UNLICENSE file for details.

"""
========
Yatiblog
========

Yatiblog provides a set of utility tools to generate HTML articles/blogs/sites
from source files.

"""

import atexit
import sys

from cStringIO import StringIO
from datetime import datetime
from os import chdir, getcwd, environ, listdir, mkdir, remove as rm, stat, walk
from os.path import (
    abspath, basename, dirname, exists, join as join_path, isfile, isdir,
    realpath, splitext
    )

from optparse import OptionParser
from pickle import load as load_pickle, dump as dump_pickle
from re import compile
from tokenize import generate_tokens, COMMENT, STRING, INDENT, NEWLINE, NL

from genshi.template import MarkupTemplate, NewTextTemplate as TextTemplate
from pygments import highlight
from pygments.lexers import get_lexer_by_name
from yaml import safe_load as load_yaml

from tavutil.env import run_command
from tavutil.scm import SCMConfig
from yatiblog.rst import render_rst, SYNTAX_FORMATTER

# ------------------------------------------------------------------------------
# Some Constants
# ------------------------------------------------------------------------------

LINE = '-' * 78
MORE_LINE = '\n.. more\n'

# ------------------------------------------------------------------------------
# Utility Functions
# ------------------------------------------------------------------------------

docstring_regex = compile(r'[ru]?"""[^"\\]*(?:(?:\\.|"(?!""))[^"\\]*)*"""')
find_include_refs = compile(r'\.\. include::\s*(.*)\s*\n').findall
match_non_whitespace = compile(r'\S').match
match_yaml_frontmatter = compile('^---\s*\n((?:.|\n)+?)\n---\s*\n').match
replace_yaml_frontmatter = compile('^---\s*\n(?:(?:.|\n)+?)\n---\s*\n').sub

def strip_leading_indent(text):
    """
    Strip common leading indentation from a piece of ``text``.

      >>> indented_text = \"""
      ...
      ...     first indented line
      ...     second indented line
      ...
      ...     \"""

      >>> print strip_leading_indent(indented_text)
      first indented line
      second indented line

    """

    lines = text.split('\n')

    if len(lines) == 1:
        return lines[0].lstrip()

    last_line = lines.pop()

    if match_non_whitespace(last_line):
        raise ValueError("Last line contains non-whitespace!")

    indent = len(last_line)
    return '\n'.join(line[indent:] for line in lines).strip()

def replace_python_docstrings(source, non_code=(COMMENT, NEWLINE, NL)):
    """Replace Python docstrings as # comment lines."""

    comments = []; out = comments.append
    prev = code = None

    for token in generate_tokens(StringIO(source).readline):
        type = token[0]
        handle = None
        if type == STRING:
            if code:
                if prev and prev[0] == INDENT:
                    handle = 1
            else:
                handle = 1
        if type not in non_code:
            code = 1
        if handle:
            docstring = token[1]
            if docstring.endswith('"""') or docstring.endswith("'''"):
                n = 3
            elif docstring.endswith('"') or docstring.endswith("'"):
                n = 1
            if docstring.startswith('r') or docstring.startswith('u'):
                docstring = docstring[n+1:-n]
            else:
                docstring = docstring[n:-n]
            out((
                token[2], token[3], (''.join(
                '\n# ' + line
                for line in strip_leading_indent(docstring).strip().splitlines()
                ) + '\n# <yatiblog.comment>\n').strip()))
        prev = token

    source_lines = source.splitlines()
    result = []; out = result.append
    prev_row, prev_col = 1, 0

    for (start_row, start_col), (end_row, end_col), comment in comments:
        if prev_row == start_row:
            block = source_lines[prev_row-1][start_col:end_col]
        else:
            start = source_lines[prev_row-1][prev_col:]
            end = source_lines[start_row-1][:start_col]
            if prev_row == (start_row - 1):
                block = '\n'.join([start, end])
            else:
                block = '\n'.join([start] + source_lines[prev_row:start_row-1] + [end])
        out(block)
        out(comment)
        prev_row, prev_col = end_row, end_col

    out('\n')
    out('\n'.join(source_lines[prev_row:]))
    return ''.join(result)

def get_git_info(filename):
    """Extract info from the Git repository."""

    environ['TZ'] = 'UTC'
    git_info = run_command(['git', 'log', '--pretty=raw', '--', filename])

    info = {'__git__': False}

    if (not git_info) or git_info.startswith('fatal:'):
        info['__updated__'] = datetime.utcfromtimestamp(
            stat(filename).st_mtime
            )
        return info

    info['__git__'] = True

    for line in git_info.splitlines():
        if line.startswith('author'):
            email, timestamp, tz = line.split()[-3:]
            email = email.lstrip('<').rstrip('>')
            if '(' in email:
                email = email.split('(')[0].strip()
            info['__by__'] = email
            info['__updated__'] = datetime.utcfromtimestamp(float(timestamp))
            break

    return info

def listfiles(path):
    """Return a list of all non-hidden files in a directory."""
    return [
        filename
        for filename in listdir(path)
        if (not filename.startswith('.'))
        ]

def load_layout(name, path, layouts, deps=None):
    """Load the given layout template."""

    template_path = join_path(path, '_layouts', name + '.genshi')
    template_file = open(template_path, 'rb')
    content = template_file.read()
    template_file.close()

    env = {}
    front_matter = match_yaml_frontmatter(content)

    if front_matter:
        env = load_yaml(front_matter.group(1))
        layout = env.pop('layout', None)
        if layout:
            if layout not in layouts:
                load_layout(layout, path, layouts)
            deps = layouts[layout]['__deps__']
            if deps:
                deps = [layout] + deps
            else:
                deps = [layout]
        content = replace_yaml_frontmatter('', content)

    if env.get('text_template'):
        try:
            template = TextTemplate(content, encoding='utf-8')
        except Exception:
            print "Error parsing template:", name
            raise
    else:
        try:
            template = MarkupTemplate(content, encoding='utf-8')
        except Exception:
            print "Error parsing template:", name
            raise

    layouts[name] = {
        '__deps__': deps,
        '__env__': env,
        '__mtime__': stat(template_path).st_mtime,
        '__name__': name,
        '__path__': template_path,
        '__template__': template,
        }

# Define the mappings for the supported programming languages.
PROGLANGS = {
    '.coffee': ['coffeescript', '#', None],
    '.el': ['scheme', ';;', None],
    '.go': ['go', '//', None],
    '.js': ['javascript', '//', None],
    '.py': ['python', '#', replace_python_docstrings],
    '.pyx': ['cython', '#', None],
    '.rb': ['ruby', '#', None],
    '.sh': ['sh', '#', None]
    }

for lang_settings in PROGLANGS.values():
    comment_symbol = lang_settings[1]
    lang_settings.extend([
        compile(r'^\s*' + comment_symbol + r'\s?'),
        '\n' + comment_symbol + ' YATIBLOG-DIVIDER\n',
        compile('<span class="c1?">'+comment_symbol+r' YATIBLOG-DIVIDER<\/span>'),
        '\n\n.. break:: YATIBLOG-DIVIDER\n\n',
        compile('<hr class="YATIBLOG-DIVIDER" />')
        ])

del comment_symbol, lang_settings

SHEBANGS = [
    ('bash', '.sh'),
    ('node', '.js'),
    ('python', '.py'),
    ('ruby', '.rb')
    ]

# ------------------------------------------------------------------------------
# Our Main Script Function
# ------------------------------------------------------------------------------

def main(argv=None):

    argv = argv or sys.argv[1:]
    op = OptionParser(
        usage="Usage: %prog [options] [path/to/source/directory]"
        )

    op.add_option('-d', dest='data_file', default='.articlestore',
                  help="Set the path for a data file (default: .articlestore)")

    op.add_option('-o', dest='output_directory', default='website',
                  help="Set the output directory for files (default: website)")

    op.add_option('-p', dest='package', default='',
                  help="Generate documentation for a Python package (optional)")

    op.add_option('--clean', dest='clean', default=False, action='store_true',
                  help="Flag to remove all generated output files")

    op.add_option('--force', dest='force', default=False, action='store_true',
                  help="Flag to force regeneration of all files")

    op.add_option('--quiet', dest='quiet', default=False, action='store_true',
                  help="Flag to suppress output")

    try:
        options, args = op.parse_args(argv)
    except SystemExit:
        return

    # Normalise various options and load from the config file.
    if args:
        source_directory = args[0]
        source_directory_specified = True
    else:
        source_directory = getcwd()
        source_directory_specified = False

    source_directory = abspath(source_directory)
    chdir(source_directory)

    if not isdir(source_directory):
        raise IOError("%r is not a directory!" % source_directory)

    config_file = join_path(source_directory, 'yatiblog.conf')

    if isfile(config_file):
        config_file_obj = open(config_file, 'rb')
        config_data = config_file_obj.read()
        config_file_obj.close()
        config = load_yaml(config_data)
    elif not source_directory_specified:
        raise IOError("Couldn't find: %s" % config_file)
    else:
        config = {}

    index_pages = config.pop('index_pages', [])
    if not isinstance(index_pages, list):
        raise ValueError("The 'index_pages' config value is not a list!")

    index_pages = dict(
        (index_page.keys()[0], index_page.values()[0])
        for index_page in index_pages
        )

    output_directory = join_path(source_directory, options.output_directory.rstrip('/'))
    if not isdir(output_directory):
        if not exists(output_directory):
            mkdir(output_directory)
        else:
            raise IOError("%r is not a directory!" % output_directory)

    code_pages = config.pop('code_pages', {})

    if code_pages:

        code_layout = code_pages['layout']
        code_paths = code_pages['paths']
        code_files = {}

        git_root = realpath(SCMConfig().root)

        for output_filename, input_pattern in code_paths.items():

            ignore_pattern = None
            if isinstance(input_pattern, dict):
                definition = input_pattern
                input_pattern = definition['pattern']
                if 'ignore' in definition:
                    ignore_pattern = definition['ignore']

            files = run_command(['git', 'ls-files', input_pattern], cwd=git_root)
            files = filter(None, files.splitlines())

            if ignore_pattern is not None:
                ignore_files = run_command(
                    ['git', 'ls-files', ignore_pattern], cwd=git_root
                    )
                for file in ignore_files.splitlines():
                    if file in files:
                        files.remove(file)

            if '%' in output_filename:
                output_pattern = True
            else:
                output_pattern = False

            for file in files:
                directory = basename(dirname(file))
                filename, ext = splitext(basename(file))
                if output_pattern:
                    dest = output_filename % {
                        'dir':directory, 'filename':filename, 'ext':ext
                        }
                else:
                    dest = output_filename
                code_files[
                    join_path(output_directory, dest + '.html')
                    ] = [file, join_path(git_root, file)]

    else:
        code_files = {}
        code_layout = None

    verbose = not options.quiet

    # See if there's a persistent data file to read from.
    data_file = join_path(source_directory, options.data_file)
    if isfile(data_file):
        data_file_obj = open(data_file, 'rb')
        data_dict = load_pickle(data_file_obj)
        data_file_obj.close()
    else:
        data_dict = {}

    # Persist the data file to disk.
    def persist_data_file():
        if data_file:
            data_file_obj = open(data_file, 'wb')
            dump_pickle(data_dict, data_file_obj)
            data_file_obj.close()

    atexit.register(persist_data_file)

    # Figure out what the generated files would be.
    source_files = [
        file for file in listfiles(source_directory) if file.endswith('.txt')
        ]

    generated_files = [
        join_path(output_directory, splitext(file)[0] + '.html')
        for file in source_files
        ]

    index_files = [join_path(output_directory, index) for index in index_pages]

    # Handle --clean support.
    if options.clean:
        for file in generated_files + index_files + [data_file] + code_files.keys():
            if isfile(file):
                if verbose:
                    print "Removing: %s" % file
                rm(file)
        data_dict.clear()
        sys.exit()

    # Figure out layout dependencies for the source .txt files.
    layouts = {}
    sources = {}

    def init_rst_source(source_file, destname=None):

        source_path = join_path(source_directory, source_file)
        source_file_obj = open(source_path, 'rb')
        content = source_file_obj.read()
        source_file_obj.close()

        if not content.startswith('---'):
            return

        filebase, filetype = splitext(source_file)
        filebase = filebase.lower()

        env = load_yaml(match_yaml_frontmatter(content).group(1))
        layout = env.pop('layout')

        if layout not in layouts:
            load_layout(layout, source_directory, layouts)

        content = replace_yaml_frontmatter('', content)

        if MORE_LINE in content:
            lead = content.split(MORE_LINE)[0]
            content = content.replace(MORE_LINE, '')
        else:
            lead = content

        if destname:
            destname = join_path(output_directory, destname)
        else:
            destname = join_path(output_directory, filebase + '.html')

        sources[source_file] = {
            '__content__': content,
            '__deps__': find_include_refs(content),
            '__env__': env,
            '__genfile__': destname,
            '__id__': source_file,
            '__layout__': layout,
            '__lead__': lead,
            '__mtime__': stat(source_path).st_mtime,
            '__name__': basename(destname), # filebase,
            '__outdir__': output_directory,
            '__path__': source_path,
            '__rst__': True,
            '__type__': 'text',
            '__filetype__': filetype
            }

    for source_file in source_files:
        init_rst_source(source_file)

    # And likewise for any source code files.
    def init_rst_source_code(relative_source_path, source_path, destname):

        source_file_obj = open(source_path, 'rb')
        content = source_file_obj.read()
        source_file_obj.close()

        filebase, filetype = splitext(basename(source_path))
        filebase = filebase.lower()

        if not filetype:
            if content.startswith('#!'):
                content = content.split('\n', 1)
                if len(content) == 2:
                    shebang, content = content
                else:
                    shebang = content[0]
                    content = ''
                for interp, ext in SHEBANGS:
                    if interp in shebang:
                        filetype = ext
                        break
            if not filetype:
                raise ValueError("Unknown file type: %s" % source_path)

        sources[source_path] = {
            '__content__': content,
            '__deps__': [],
            '__env__': {'title': filebase},
            '__genfile__': destname,
            '__gitpath__': relative_source_path,
            '__id__': source_path,
            '__layout__': code_layout,
            '__lead__': '',
            '__mtime__': stat(source_path).st_mtime,
            '__name__': basename(destname), # filebase,
            '__outdir__': output_directory,
            '__path__': source_path,
            '__rst__': True,
            '__type__': 'code',
            '__filetype__': filetype
            }

    if code_layout and code_layout not in layouts:
        load_layout(code_layout, source_directory, layouts)

    for destname, (relative_source_path, source_path) in code_files.items():
        init_rst_source_code(relative_source_path, source_path, destname)

    # And likewise for the ``index_pages``.
    render_last = set()

    for index_page, index_source in index_pages.items():
        layout, filetype = splitext(index_source)
        if filetype == '.genshi':
            if layout not in layouts:
                load_layout(layout, source_directory, layouts)
            source_path = join_path(source_directory, '_layouts', index_source)
            sources[index_source] = {
                '__content__': '',
                '__deps__': [],
                '__env__': {},
                '__genfile__': join_path(output_directory, index_page),
                '__id__': index_source,
                '__layout__': layout,
                '__lead__': '',
                '__mtime__': stat(source_path).st_mtime,
                '__name__': basename(index_page),
                '__outdir__': output_directory,
                '__path__': source_path,
                '__rst__': False,
                '__type__': 'index',
                '__filetype__': 'genshi'
                }
        else:
            init_rst_source(index_source, index_page)
        render_last.add(index_source)

    # Update the envs for all the source files.
    for source in sources:
        info = sources[source]
        layout = info['__layout__']
        layout_info = layouts[layout]
        if layout_info['__deps__']:
            for dep_layout in reversed(layout_info['__deps__']):
                info.update(layouts[dep_layout]['__env__'])
        info.update(layouts[layout]['__env__'])
        info.update(get_git_info(info['__path__']))
        info.update(info.pop('__env__'))

    # Figure out which files to regenerate.
    if not options.force:

        no_regen = set()
        for source in sources:

            info = sources[source]
            try:
                gen_mtime = stat(info['__genfile__']).st_mtime
            except:
                continue

            dirty = False
            if gen_mtime < info['__mtime__']:
                dirty = True

            layout = info['__layout__']
            layout_info = layouts[layout]
            if layout_info['__deps__']:
                layout_chain = [layout] + layout_info['__deps__']
            else:
                layout_chain = [layout]

            for layout in layout_chain:
                if gen_mtime < layouts[layout]['__mtime__']:
                    dirty = True
                    break

            for dep in info['__deps__']:
                dep_mtime = stat(join_path(source_directory, dep)).st_mtime
                if gen_mtime < dep_mtime:
                    dirty = True
                    break

            if not dirty:
                no_regen.add(source)

        for source in no_regen:
            if source in render_last:
                continue
            del sources[source]

        remaining = set(sources.keys())
        if remaining == render_last:
            for source in remaining.intersection(no_regen):
                del sources[source]

    BLANK_CODE_LINE = '<div class="syntax"><pre><div class="syntax"><pre></pre></div>'

    # Regenerate!
    items = sorted(sources.items(), key=lambda x: x[1]['__rst__'] == False)

    for source, source_info in items:

        info = config.copy()
        info.update(source_info)

        if verbose:
            print
            print LINE
            print 'Converting: [%s] %s' % (info['__type__'], info['__path__'])
            print LINE
            print

        if info['__type__'] == 'code':

            content = info['__content__']
            conf = PROGLANGS[info['__filetype__']]
            if conf[2]:
                content = conf[2](content)
            comment_matcher = conf[3]

            lines = content.split('\n')
            include_section = None

            if lines and lines[0].startswith('#!'):
                lines.pop(0)

            sections = []; new_section = sections.append
            docs_text = []; docs_out = docs_text.append
            code_text = []; code_out = code_text.append

            for line in lines:
                if comment_matcher.match(line):
                    line = comment_matcher.sub('', line)
                    if line == '<yatiblog.comment>':
                        include_section = 1
                    else:
                        docs_out(line)
                else:
                    if not line.strip():
                        if docs_text and not include_section:
                            last_line = docs_text[-1].strip()
                            if last_line:
                                last_line_char = last_line[0]
                                for char in last_line:
                                    if char != last_line_char:
                                        break
                                else:
                                    include_section = 1
                    else:
                        if docs_text:
                            include_section = 1
                    if docs_text:
                        if include_section:
                            new_section({
                                'docs_text': '\n'.join(docs_text) + '\n',
                                'code_text': '\n'.join(code_text)
                                })
                            docs_text[:] = []
                            code_text[:] = []
                            include_section = None
                        else:
                            docs_text[:] = []
                        code_out(line)
                    else:
                        code_out(line)

            new_section({'docs_text': '', 'code_text': '\n'.join(code_text)})

            docs = conf[6].join(part['docs_text'] for part in sections)
            code = conf[4].join(part['code_text'] for part in sections)

            docs_html, props = render_rst(docs, with_props=1)
            if ('title' in props) and props['title']:
                info['title'] = props['title']

            code = code.replace('\t', '    ')
            code_html = highlight(code, get_lexer_by_name(conf[0]), SYNTAX_FORMATTER)

            docs_split = conf[7].split(docs_html)
            code_split = conf[5].split(code_html)
            output = info['__output__'] = []
            out = output.append

            if docs_split and docs_split[0]:
                diff = 0
                docs_split.insert(0, u'')
            else:
                diff = 1

            last = len(docs_split) - 2
            for i in range(last + 1):
                code = code_split[i+diff].split(u'<br/>')
                while (code and code[0] == ''):
                    code.pop(0)
                while (code and code[-1] == ''):
                    code.pop()
                code = u'<br />'.join(code)
                if code:
                    if i == last:
                        code = u'<div class="syntax"><pre>' + code
                    else:
                        code = u'<div class="syntax"><pre>' + code + "</pre></div>"
                out((docs_split[i], code))

            while output and output[0][1] == BLANK_CODE_LINE:
                if not output[0][0]:
                    output.pop(0)

        elif info['__rst__']:
            with_props = info.get('with_props', False)
            if with_props:
                output, props = render_rst(info['__content__'], with_props=1)
                if ('title' in props) and props['title']:
                    info['title'] = props['title']
                info['__output__'] = output
            else:
                output = info['__output__'] = render_rst(info['__content__'])

            if info['__lead__'] == info['__content__']:
                info['__lead_output__'] = info['__output__']
            else:
                info['__lead_output__'] = render_rst(info['__lead__'])
        else:
            output = ''

        layout = info['__layout__']
        layout_info = layouts[layout]

        if layout_info['__deps__']:
            layout_chain = [layout] + layout_info['__deps__']
        else:
            layout_chain = [layout]

        for layout in layout_chain:
            template = layouts[layout]['__template__']
            output = template.generate(
                content=output,
                yatidb=data_dict,
                **info
                ).render('xhtml', encoding=None)

        if isinstance(output, unicode):
            output = output.encode('utf-8')

        data_dict[info['__name__']] = info

        output_file = open(info['__genfile__'], 'wb')
        output_file.write(output)
        output_file.close()

        if verbose:
            print 'Done!'

    sys.exit()

# PDF_COMMAND = ['prince', '--input=html', '--output=pdf'] # --no-compress
# PDF_CSS = join_path(WEBSITE_ROOT, 'static', 'css', 'print.css')
# $(documentation_pdf_files): documentation/pdf/%.pdf: documentation/article/%.html $(template) $(pdf_css)
# 	@echo "---> generating" "$@"
# 	    $(prince) $$n --style=$(pdf_css) --output=$@; \

# ------------------------------------------------------------------------------
# Run Farmer!
# ------------------------------------------------------------------------------

if __name__ == '__main__':
    main()
